----------------------------------------------------------------------
-- VGA controller 640X480 bitmap monochrome                         -- 
-- Full RTL synchronous pipelined architecture                      --
-- Project by Valerio Venturi (Italy)                               -- 
-- Date: 14/04/2011                                                 --
-- Last revision: 05/05/2011                                        --
----------------------------------------------------------------------

library IEEE;
use IEEE.std_logic_1164.all;                             -- defines std_logic types
use IEEE.STD_LOGIC_unsigned.all;
use IEEE.STD_LOGIC_arith.all;

-- global architecture 
entity vga is
  port(    clock:    in STD_LOGIC;                       -- clock input (25 MHZ)
             res:    in STD_LOGIC;                       -- reset input
            addr:   out STD_LOGIC_VECTOR(15 downto 0);   -- 16 bit refresh address bus out (to video RAM)
         data_in:    in STD_LOGIC_VECTOR(7 downto 0);    -- 8 bit input data bus (from video RAM)
            vdma:   out STD_LOGIC;                       -- video DMA request (used to halt the CPU and switch the RAM address multiplexer)  
          shf_ld:   out STD_LOGIC;                       -- shifter load
           hsync:   out STD_LOGIC;                       -- horizontal sync
           vsync:   out STD_LOGIC;                       -- vertical sync
             ven:   out STD_LOGIC;                       -- video enable (blank)
           pixel:   out STD_LOGIC                        -- pixel data
      );   
end vga;

architecture struct of vga is
constant START_VIDEO_MEM:       STD_LOGIC_VECTOR(15 downto 0) := "0001000000000000";  -- maps video memory base address to 0x1000 

signal videoen, videov, videoh: STD_LOGIC;
signal hcount, vcount:          STD_LOGIC_VECTOR(9 downto 0);
signal reset:                   STD_LOGIC;
signal i_pixel:                 STD_LOGIC_VECTOR(7 downto 0);   -- internal pixel data
signal vid_ld:                  STD_LOGIC_VECTOR(2 downto 0);               
signal maddr:                   STD_LOGIC_VECTOR(15 downto 0);   
signal ma_upd:                  STD_LOGIC;
signal ma_clr:                  STD_LOGIC;            

begin

-- reset is active high
reset <= not res;

-- horizontal counter
hcounter: process (clock, reset)
begin
  if reset='1' then 
    hcount <= (others => '0');
  else 
    if (clock'event and clock='1') then 
      if hcount = 799 then 
        hcount <= (others => '0');
      else 
        hcount <= hcount + 1;
      end if;
    end if;
  end if;
end process;

-- horizontal active portion (0 to 639)
process (hcount)
begin
  videoh <= '1';
  if hcount > 639 then 
    videoh <= '0';
  end if;
end process;

-- vertical counter
vcounter: process (clock, reset)
begin
  if reset='1' then 
    vcount <= (others => '0');
  else 
    if (clock'event and clock='1') then 
      if hcount = 699 then 
        if vcount = 524 then 
          vcount <= (others => '0');
        else 
          vcount <= vcount + 1;
        end if;
      end if;
    end if;
  end if;
end process;

-- vertical active portion (0 to 479)
process (vcount)
begin
  videov <= '1';
  if vcount > 479 then 
    videov <= '0';
  end if;
end process;

-- VGA sync generation
sync: process (clock, reset)
begin
  if reset='1' then 
    hsync <= '0';
    vsync <= '0';
    ma_clr <= '1';
  else 
    if (clock'event and clock='1') then 
      if (hcount <= 755 and hcount >= 659) then 
        hsync <= '0';
      else 
        hsync <= '1';
      end if;
      if (vcount <= 494 and vcount >= 493) then 
        vsync <= '0';
        ma_clr <= '1';
      else 
        vsync <= '1';
        ma_clr <= '0';
      end if;
    end if;
  end if;
end process;

-- video ram scanner
vram: process(clock)
begin
  if(clock'event and clock = '1')then 
    if ma_clr = '1' then
      maddr <= START_VIDEO_MEM;
    else
      if ma_upd = '1' then
        if maddr < 42495 then
          maddr <= maddr +1;
        else  
          maddr <= START_VIDEO_MEM;
        end if;
      else
        maddr <= maddr;
      end if;
    end if;        
  end if;
end process;  

-- video shifter
-- one byte read from video RAM correspond to eight pixel shifted out starting from MSB bit
video: process (clock, reset)
begin
  if reset = '1' then
    vdma <= '1';
    shf_ld <= '0';
    ma_upd <= '0';
    i_pixel <= (others => '0');
    vid_ld <= (others => '1');
  else
    if (clock'event and clock = '1') then 
      if videoen = '1' then
        vid_ld <= vid_ld +1;
        if vid_ld = 7 then
          vdma <= '0';
        else
          vdma <= '1';
        end if;
        if vid_ld = 0 then    
          shf_ld <= '1';
          i_pixel <= data_in;
        else
          shf_ld <= '0';
          i_pixel <= i_pixel(i_pixel'left-1 downto i_pixel'right) & '0';
        end if;
        if vid_ld = 1 then
          ma_upd <= '1';
        else
          ma_upd <= '0';
        end if;    
      else
        vdma <= '1';
        shf_ld <= '0';
        ma_upd <= '0';
        i_pixel <= (others => '0');
      end if; 
    end if;
  end if;    
end process;

videoen <= videoh and videov;
ven <= videoen;
pixel <= i_pixel(7);
addr <= maddr;
   
end struct;    
  


